#include "udfs/commands_crud/bson_update_document--0.111-0.sql"
#include "udfs/commands_diagnostic/bson_get_rewritten_sql--0.111-0.sql"

-- New optimized group aggregate implementations
#include "udfs/aggregation/group_aggregates_support--0.111-0.sql"
#include "udfs/aggregation/group_aggregates--0.111-0.sql"

#include "udfs/query/bson_orderby--0.111-0.sql"
#include "operators/bson_dollar_operators--0.111-0.sql"
#include "udfs/query/bson_stats_project--0.111-0.sql"

#include "udfs/commands_crud/bulkWrite--0.111-0.sql"

-- RBAC schema conditional function definitions
#include "udfs/commands_crud/cursor_functions--0.111-0.sql"
#include "udfs/commands_crud/delete--0.111-0.sql"
#include "udfs/commands_crud/find_and_modify--0.111-0.sql"
#include "udfs/commands_crud/insert--0.111-0.sql"
#include "udfs/commands_crud/insert_one_helper--0.111-0.sql"
#include "udfs/commands_crud/query_cursors_aggregate--0.111-0.sql"
#include "udfs/commands_crud/query_cursors_single_page--0.111-0.sql"
#include "udfs/commands_crud/update--0.111-0.sql"
#include "udfs/commands_diagnostic/coll_stats--0.111-0.sql"
#include "udfs/commands_diagnostic/current_op--0.111-0.sql"
#include "udfs/commands_diagnostic/db_stats--0.111-0.sql"
#include "udfs/commands_diagnostic/index_stats--0.111-0.sql"
#include "udfs/commands_diagnostic/validate--0.111-0.sql"
#include "udfs/index_mgmt/create_builtin_id_index--0.111-0.sql"
#include "udfs/index_mgmt/create_index_background--0.111-0.sql"
#include "udfs/index_mgmt/create_indexes_non_concurrently--0.111-0.sql"
#include "udfs/index_mgmt/drop_indexes--0.111-0.sql"
#include "udfs/index_mgmt/record_id_index--0.111-0.sql"
#include "udfs/metadata/collection--0.111-0.sql"
#include "udfs/metadata/empty_data_table--0.111-0.sql"
#include "udfs/metadata/list_databases--0.111-0.sql"
#include "udfs/schema_mgmt/create_collection--0.111-0.sql"
#include "udfs/schema_mgmt/create_collection_view--0.111-0.sql"
#include "udfs/schema_mgmt/drop_collection--0.111-0.sql"
#include "udfs/schema_mgmt/drop_database--0.111-0.sql"
#include "udfs/schema_mgmt/rename_collection--0.111-0.sql"
