/*-------------------------------------------------------------------------
 * Copyright (c) Microsoft Corporation.  All rights reserved.
 *
 * documentdb_tests/src/commands/mod.rs
 *
 *-------------------------------------------------------------------------
 */

pub mod aggregate;
pub mod coll_stats;
pub mod collection_cmd;
pub mod compact;
pub mod constant;
pub mod count;
pub mod current_op;
pub mod cursor;
pub mod delete;
pub mod distinct;
pub mod explain;
pub mod find;
pub mod find_and_modify;
pub mod indexing;
pub mod insert;
pub mod killop;
pub mod list_collections;
pub mod list_databases;
pub mod multi_connect;
pub mod rbac_builtin_read_any_database_tests;
pub mod rename_collection;
pub mod session;
pub mod text_search;
pub mod transaction;
pub mod unsupported_commands;
pub mod update;
pub mod users;
pub mod validate_cmd;
